@@ -4,8 +4,62 @@ Definitions of various types of fields. Supports JSON draft4 types.
 import re
 from collections import OrderedDict
 from functools import reduce
+from decimal import Decimal, getcontext, InvalidOperation
 
-from typedpy.structures import Field, Structure, TypedField, ClassReference
+from typedpy.structures import Field, Structure, TypedField, ClassReference, StructMeta
+
+
+def _map_to_field(item):
+    if isinstance(item, StructMeta) and not isinstance(item, Field):
+        return ClassReference(item)
+    if item in [None, ''] or isinstance(item, Field):
+        return item
+    elif Field in getattr(item, '__mro__', []):
+        return item()
+    else:
+        raise TypeError("Expected a Field/Structure class or Field instance")
+
+
+class StructureReference(Field):
+    """
+    A Field that is an embedded structure within other structure. Allows to create hierarchy.
+    This is useful if you want to inline your Structure, as opposed to create an explicit
+    class for it.
+    All the arguments are passed as attributes of the structure. Example:
+
+    .. code-block:: python
+
+        StructureReference(
+            _additionalProperties = False,
+            id = String,
+            name = String
+            age = AnyOf[PositiveInt, PositiveFloat]
+        )
+
+    """
+    counter = 0
+
+    def __init__(self, **kwargs):
+        classname = "StructureReference_" + str(StructureReference.counter)
+        StructureReference.counter += 1
+
+        self._newclass = type(classname, (Structure,), kwargs)
+        super().__init__(kwargs)
+
+    def __set__(self, instance, value):
+        if not isinstance(value, dict):
+            raise TypeError("{}: Expected a dictionary".format(self._name))
+        newval = self._newclass(**value)
+        super().__set__(instance, newval)
+
+    def __str__(self):
+        props = []
+        for k, val in sorted(self._newclass.__dict__.items()):
+            if val is not None and not k.startswith('_'):
+                props.append('{} = {}'.format(k, str(val)))
+
+        propst = '. Properties: {}'.format(', '.join(props)) if props else ''
+        return '<Structure{}>'.format(propst)
 
 
 class ImmutableField(Field):
@@ -37,28 +91,37 @@ class Number(Field):
         self.exclusiveMaximum = exclusiveMaximum
         super().__init__(*args, **kwargs)
 
-    def __set__(self, instance, value):
+    @staticmethod
+    def _validate_static(self, value):
         def is_number(val):
-            return isinstance(val, (float, int))
+            return isinstance(val, (float, int, Decimal))
 
-        if not isinstance(value, float) and not isinstance(value, int):
-            raise TypeError("{}: Expected a number".format(self._name))
+        def err_prefix():
+            return "{}: ".format(self._name) if self._name else ""
+        if not is_number(value):
+            raise TypeError("{}Expected a number".format(err_prefix()))
         if isinstance(self.multiplesOf, float) and \
-                        int(value / self.multiplesOf) != value / self.multiplesOf or \
-                        isinstance(self.multiplesOf, int) and value % self.multiplesOf:
-            raise ValueError("{}: Expected a a multiple of {}".format(
-                self._name, self.multiplesOf))
+                int(value / self.multiplesOf) != value / self.multiplesOf or \
+                isinstance(self.multiplesOf, int) and value % self.multiplesOf:
+            raise ValueError("{}Expected a a multiple of {}".format(
+                err_prefix(), self.multiplesOf))
         if (is_number(self.minimum)) and self.minimum > value:
-            raise ValueError("{}: Expected a minimum of {}".format(
-                self._name, self.minimum))
+            raise ValueError("{}Expected a minimum of {}".format(
+                err_prefix(), self.minimum))
         if is_number(self.maximum):
             if self.exclusiveMaximum and self.maximum == value:
-                raise ValueError("{}: Expected a maxmimum of less than {}".format(
-                    self._name, self.maximum))
+                raise ValueError("{}Expected a maxmimum of less than {}".format(
+                    err_prefix(), self.maximum))
             else:
                 if self.maximum < value:
-                    raise ValueError("{}: Expected a maxmimum of {}".format(
-                        self._name, self.maximum))
+                    raise ValueError("{}Expected a maxmimum of {}".format(
+                        err_prefix(), self.maximum))
+
+    def _validate(self, value):
+        Number._validate_static(self, value)
+
+    def __set__(self, instance, value):
+        self._validate(value)
         super().__set__(instance, value)
 
 
@@ -68,6 +131,26 @@ class Integer(TypedField, Number):
     """
     _ty = int
 
+    def _validate(self, value):
+        super()._validate(value)
+        Number._validate_static(self, value)
+
+
+class DecimalNumber(Number):
+    """
+    An extension of :class:`Number` for a Decimal. Accepts anything that can be converted to a Decimal
+    """
+
+    def __set__(self, instance, value):
+        try:
+            value = Decimal(value)
+        except TypeError as ex:
+            raise TypeError("{}: {}".format(self._name, ex.args[0]))
+        except InvalidOperation as ex:
+            raise ValueError("{}: {}".format(self._name, ex.args[0]))
+
+        super().__set__(instance, value)
+
 
 class String(TypedField):
     """
@@ -93,28 +176,61 @@ class String(TypedField):
             self._compiled_pattern = re.compile(self.pattern)
         super().__init__(*args, **kwargs)
 
-    def __set__(self, instance, value):
-        if not isinstance(value, str):
-            raise TypeError("{}: Expected a string".format(self._name))
-        if self.maxLength is not None and len(value) > self.maxLength:
-            raise ValueError("{}: Expected a maxmimum length of {}".format(
-                self._name, self.maxLength))
-        if self.minLength is not None and len(value) < self.minLength:
-            raise ValueError("{}: Expected a minimum length of {}".format(
-                self._name, self.minLength))
-        if self.pattern is not None and not self._compiled_pattern.match(value):
-            raise ValueError('{}: Does not match regular expression: {}'.format(
-                self._name, self.pattern))
+    def _validate(self, value):
+        String._validate_static(self, value)
 
+    @staticmethod
+    def _validate_static(self, value):
+        def err_prefix():
+            return "{}: ".format(self._name) if self._name else ""
+
+        if not isinstance(value, str):
+            raise TypeError("{}Expected a string".format(err_prefix()))
+        if self.maxLength is not None and len(value) > self.maxLength:
+            raise ValueError("{}Expected a maxmimum length of {}".format(
+                err_prefix(), self.maxLength))
+        if self.minLength is not None and len(value) < self.minLength:
+            raise ValueError("{}Expected a minimum length of {}".format(
+                err_prefix(), self.minLength))
+        if self.pattern is not None and not self._compiled_pattern.match(value):
+            raise ValueError('{}Does not match regular expression: "{}"'.format(
+                err_prefix(), self.pattern))
+
+    def __set__(self, instance, value):
+        self._validate(value)
         super().__set__(instance, value)
 
 
+class Anything(Field):
+    """
+    A field that can contain anything (similar to "any" in Typescript).
+    Example:
+
+    .. code-block:: python
+
+        class Foo(Structure):
+            i = Integer
+            some_content = Anything
+
+        # now we can assign anything to some_content property:
+        Foo(i=5, some_content = "whatever")
+        Foo(i=5, some_content = [1,2,3])
+        Foo(i=5, some_content = Bar())
+
+    """
+    pass
+
+
 class Float(TypedField, Number):
     """
     An extension of :class:`Number` for a float
     """
     _ty = float
 
+    def _validate(self, value):
+        super()._validate(value)
+        Number._validate_static(self, value)
+
 
 class Boolean(TypedField):
     """
@@ -127,6 +243,7 @@ class Positive(Number):
     """
     An extension of :class:`Number`. Requires the number to be positive
     """
+
     def __set__(self, instance, value):
         if value <= 0:
             raise ValueError('{}: Must be positive'.format(self._name))
@@ -174,6 +291,7 @@ class _ListStruct(list):
         copied = self.copy()
         copied.extend(value)
         setattr(self._instance, getattr(self._array, '_name', None), copied)
+        setattr(self._instance, getattr(self._array, '_name', None), copied)
 
     def insert(self, index: int, value):
         copied = self.copy()
@@ -242,6 +360,11 @@ class _CollectionMeta(type):
         return cls(items=validate_and_get_field(item))
 
 
+class _EnumMeta(type):
+    def __getitem__(cls, values):
+        return cls(values=list(values))
+
+
 class _JSONSchemaDraft4ReuseMeta(type):
     def __getitem__(cls, item):
         def validate_and_get_field(val):
@@ -284,8 +407,9 @@ class Set(SizedCollection, TypedField, metaclass=_CollectionMeta):
             minimal size
         maxItems(int): optional
             maximal size
-        items(:class:`Field`): optional
-            The type of the content
+        items(:class:`Field` or :class:`Structure`): optional
+            The type of the content, can be a predefined :class:`Structure` or
+            :class:`Field`
 
     Examples:
 
@@ -294,18 +418,21 @@ class Set(SizedCollection, TypedField, metaclass=_CollectionMeta):
         Set[String]
         Set(items=Integer(maximum=10), maxItems = 10)
 
+        # let's assume we defined a Structure 'Person', then we can use it too:
+        Set[Person]
+
 
     """
     _ty = set
 
     def __init__(self, *args, items=None,
                  **kwargs):
-        if items is None or isinstance(items, Field):
-            self.items = items
-        elif Field in getattr(items, '__mro__', []):
-            self.items = items()
-        else:
-            raise TypeError("Expected a Field class or instance")
+        self.items = _map_to_field(items)
+
+        if isinstance(self.items, TypedField) and not \
+                getattr(getattr(self.items, '_ty'), '__hash__'):
+            raise TypeError("Set element of type {} is not hashable".format(
+                getattr(self.items, '_ty')))
         super().__init__(*args, **kwargs)
 
     def __set__(self, instance, value):
@@ -332,13 +459,18 @@ class Map(SizedCollection, TypedField, metaclass=_CollectionMeta):
             minimal size
         maxItems(int): optional
             maximal size
-        items(tuple of 2 :class:`Field` elements): optional
+        items(tuple of 2 :class:`Field` or :class:`Structure`  elements): optional
             The first element is the Field for keys, the second is for values.
             Examples:
 
             .. code-block:: python
 
                 age_by_name = Map[String, PositiveInt]
+                # Let's assume we defined a Structure "Person"
+                person_by_id = Map[String, Person]
+                # even Structure reference is supported for keys!
+                id_by_person = Map[Person, String]
+                id_by_person = Map[Person, String]
 
     """
 
@@ -353,12 +485,11 @@ class Map(SizedCollection, TypedField, metaclass=_CollectionMeta):
         else:
             self.items = []
             for item in items:
-                if isinstance(item, Field):
-                    self.items.append(item)
-                elif Field in item.__mro__:
-                    self.items.append(item())
-                else:
-                    raise TypeError("Expected a Field class or instance")
+                self.items.append(_map_to_field(item))
+            key_field = self.items[0]
+            if isinstance(key_field, TypedField) and not getattr(getattr(key_field, '_ty'), '__hash__'):
+                raise TypeError("Key field of type {}, with underlying type of {} is not hashable".format(
+                    key_field, getattr(key_field, '_ty')))
         super().__init__(*args, **kwargs)
 
     def __set__(self, instance, value):
@@ -398,7 +529,7 @@ class Array(SizedCollection, TypedField, metaclass=_CollectionMeta):
         additionalItems(bool): optional
             Relevant in case items parameter is a list of Fields. Is it allowed to have additional
             elements beyond the ones defined in "items"?
-        items(a :class:`Field` or a list/tuple of :class:`Field`): optional
+        items(a :class:`Field` or :class:`Structure`, or a list/tuple of :class:`Field` or :class:`Structure`): optional
             Describes the fields of the elements.
             If a items if a :class:`Field`, then it applies to all items.
             If a items is a list, then every element in the content is expected to be
@@ -412,6 +543,9 @@ class Array(SizedCollection, TypedField, metaclass=_CollectionMeta):
                 names = Array(minItems=5, items=String)
                 my_record = Array(items=[String, Integer(minimum=5), String])
                 my_lists = Array[Array[Integer]]
+                my_structs = Array[StructureReference(a=Integer, b=Float)]
+                # Let's say we defined a Structure "Person"
+                people = Array[Person]
 
     """
     _ty = list
@@ -433,26 +567,14 @@ class Array(SizedCollection, TypedField, metaclass=_CollectionMeta):
         if isinstance(items, list):
             self.items = []
             for item in items:
-                if isinstance(item, Field):
-                    self.items.append(item)
-                elif Field in item.__mro__:
-                    self.items.append(item())
-                else:
-                    raise TypeError("Expected a Field class or instance")
+                self.items.append(_map_to_field(item))
         else:
-            self.items = items
+            self.items = _map_to_field(items)
         super().__init__(*args, **kwargs)
 
     def __set__(self, instance, value):
-        if not isinstance(value, list):
-            raise TypeError("%s: Expected %s" % (self._name, list))
+        verify_type_and_uniqueness(list, value, self._name, self.uniqueItems)
         self.validate_size(value, self._name)
-        if self.uniqueItems:
-            unique = reduce(lambda unique_vals, x: unique_vals.append(x) or
-                            unique_vals if x not in unique_vals
-                            else unique_vals, value, [])
-            if len(unique) < len(value):
-                raise ValueError("{}: Expected unique items".format(self._name))
         if self.items is not None:
             if isinstance(self.items, Field):
                 temp_st = Structure()
@@ -482,6 +604,17 @@ class Array(SizedCollection, TypedField, metaclass=_CollectionMeta):
         super().__set__(instance, _ListStruct(self, instance, value))
 
 
+def verify_type_and_uniqueness(the_type, value, name, has_unique_items):
+    if not isinstance(value, the_type):
+        raise TypeError("%s: Expected %s" % (name, the_type))
+    if has_unique_items:
+        unique = reduce(lambda unique_vals, x: unique_vals.append(x) or
+                        unique_vals if x not in unique_vals
+                        else unique_vals, value, [])
+        if len(unique) < len(value):
+            raise ValueError("{}: Expected unique items".format(name))
+
+
 class Tuple(TypedField, metaclass=_CollectionMeta):
     """
     A tuple field, supports unique items option.
@@ -492,19 +625,33 @@ class Tuple(TypedField, metaclass=_CollectionMeta):
         unqieItems(`bool`): optional
             are elements required to be unique?
 
-        items(`list`/`tuple` of :class:`Field`): optional
+        items(`list`/`tuple` of :class:`Field` or :class:`Structure`): optional
             Describes the fields of the elements.
             Every element in the content is expected to be
             of the corresponding :class:`Field` in items.
 
+
     Examples:
 
     .. code-block:: python
 
+        // a is a tuple of exactly 2 strings that are different from each other.
         a = Tuple(uniqueItems=True, items = [String, String])
+
+        // b is a tuple of 3: string, string and a number up to 10.
         b = Tuple(items = [String, String, Number(maximum=10)])
+
+        // c is a tuple of 3: integer, string, float.
         c = Tuple[Integer, String, Float]
 
+        // The following define a tuple of any number of Integers
+        d = Tuple[Integer]
+
+        // It can also contain other structures:
+        // Assume we have something like: class Foo(Structure): pass
+        // e is a tuple of any number of Integers or Foo instances
+        e = Tuple[AnyOf[Integer, Foo]]
+
     """
     _ty = tuple
 
@@ -528,36 +675,32 @@ class Tuple(TypedField, metaclass=_CollectionMeta):
                     self.items.append(item())
                 else:
                     raise TypeError("Expected a Field class or instance")
+        elif isinstance(items, (Field,)) or Field in items.__mro__:
+            self.items = [items]
         else:
-            raise TypeError("Expected a list/tuple of Fields")
+            raise TypeError("Expected a list/tuple of Fields or a single Field")
         super().__init__(*args, **kwargs)
 
     def __set__(self, instance, value):
-        if not isinstance(value, tuple):
-            raise TypeError("%s: Expected %s" % (self._name, tuple))
-        if self.uniqueItems:
-            unique = reduce(lambda unique_vals, x: unique_vals.append(x) or
-                            unique_vals if x not in unique_vals
-                            else unique_vals, value, [])
-            if len(unique) < len(value):
-                raise ValueError("{}: Expected unique items".format(self._name))
-        if len(self.items) != len(value):
+        verify_type_and_uniqueness(tuple, value, self._name, self.uniqueItems)
+        if len(self.items) != len(value) and len(self.items) > 1:
             raise ValueError("{}: Expected a tuple of length {}".format(
                 self._name, len(self.items)))
 
         temp_st = Structure()
         res = []
-        for ind, item in enumerate(self.items):
+        items = self.items if len(self.items) > 1 else self.items * len(value)
+        for ind, item in enumerate(items):
             setattr(item, '_name', self._name + "_{}".format(str(ind)))
             item.__set__(temp_st, value[ind])
             res.append(getattr(temp_st, getattr(item, '_name')))
-            res += value[len(self.items):]
+            res += value[len(items):]
         value = tuple(res)
 
         super().__set__(instance, value)
 
 
-class Enum(Field):
+class Enum(Field, metaclass=_EnumMeta):
     """
         Enum field. value can be one of predefined values
 
@@ -566,19 +709,33 @@ class Enum(Field):
                  allowed values. Can be of any type
 
     """
+
     def __init__(self, *args, values, **kwargs):
         self.values = values
         super().__init__(*args, **kwargs)
 
-    def __set__(self, instance, value):
+    def _validate(self, value):
         if value not in self.values:
             raise ValueError('{}: Must be one of {}'.format(self._name, self.values))
+
+    def __set__(self, instance, value):
+        self._validate(value)
         super().__set__(instance, value)
 
 
 class EnumString(Enum, String):
     """
-    Combination of :class:`Enum` and :class:`String`
+    Combination of :class:`Enum` and :class:`String`. This is useful if you want to further
+    limit your allowable enum values, using :class:`String` attributes, such as pattern, maxLength.
+
+    Example:
+
+    .. code-block:: python
+
+        predefined_list = ['abc', 'x', 'def', 'yy']
+
+        EnumString(values=predefined_list, minLength=3)
+
     """
     pass
 
@@ -594,6 +751,7 @@ class Sized(Field):
                 maximum length
 
     """
+
     def __init__(self, *args, maxlen, **kwargs):
         self.maxlen = maxlen
         super().__init__(*args, **kwargs)
@@ -623,16 +781,12 @@ class MultiFieldWrapper(object):
     An abstract base class for AllOf, AnyOf, OneOf, etc.
     It provides flexibility in reading the "fields" argument.
     """
+
     def __init__(self, *arg, fields, **kwargs):
         if isinstance(fields, list):
             self._fields = []
             for item in fields:
-                if isinstance(item, Field):
-                    self._fields.append(item)
-                elif Field in item.__mro__:
-                    self._fields.append(item())
-                else:
-                    raise TypeError("Expected a Field class or instance")
+                self._fields.append(_map_to_field(item))
         else:
             raise TypeError("Expected a Field class or instance")
         super().__init__(*arg, **kwargs)
@@ -656,6 +810,7 @@ class AllOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
         AllOf[Number(maximum=20, minimum=-10), Integer, Positive]
 
     """
+
     def __init__(self, fields):
         super().__init__(fields=fields)
 
@@ -684,6 +839,7 @@ class AnyOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
        AnyOf[Number(maximum=20, minimum=-10), Integer, Positive, String]
 
     """
+
     def __init__(self, fields):
         super().__init__(fields=fields)
 
@@ -699,7 +855,7 @@ class AnyOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
             except ValueError:
                 pass
         if not matched:
-            raise ValueError("{}: Did not match any field option".format(self._name))
+            raise ValueError("{}: {} Did not match any field option".format(self._name, value))
         super().__set__(instance, value)
 
     def __str__(self):
@@ -721,6 +877,7 @@ class OneOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
         OneOf[Number(maximum=20, minimum=-10), Integer, Positive, String]
 
     """
+
     def __init__(self, fields):
         super().__init__(fields=fields)
 
@@ -761,6 +918,7 @@ class NotField(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
         NotField[Positive]
 
     """
+
     def __init__(self, fields):
         super().__init__(fields=fields)
 
@@ -785,7 +943,7 @@ class NotField(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
 class ValidatedTypedField(TypedField):
 
     def __set__(self, instance, value):
-        self._validate_func(value) # pylint: disable=E1101
+        self._validate_func(value)  # pylint: disable=E1101
         super().__set__(instance, value)
 
 
@@ -816,9 +974,10 @@ def create_typed_field(classname, cls, validate_func=None):
         classname(`str`):
             the content must not match any of the fields in the lists
     """
+
     def validate_wrapper(cls, value):
         if validate_func is None:
             return
         validate_func(value)
 
-    return type(classname, (ValidatedTypedField,), {'_validate_func' :validate_wrapper, '_ty': cls})
+    return type(classname, (ValidatedTypedField,), {'_validate_func': validate_wrapper, '_ty': cls})
